/*
  Copyright (c) 2005 Jeremy Garff
  Please see documentation for licensing information.
*/

#include <avr/io.h>
#include <avr/signal.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>

#include "usbdrv/usbdrv.h"
#include "ledloadmsg.h"

#define STATE_OFF      0
#define STATE_ON       1

#define GREEN_BIT      1
#define BLUE_BIT       2
#define RED_BIT        4

#define TIMEOUT_WAIT   4800
#define WARN_ON        600 + TIMEOUT_WAIT
#define WARN_OFF       1200 + TIMEOUT_WAIT

struct color_tracker {
  uchar state;
  uchar count;
  uchar color;
  uchar ledbit;
};

#define RED            2
#define GREEN          0
#define BLUE           1

struct color_tracker track[3];

uchar initted = 0;
unsigned short gc = 0;


/*
 * Interrupt Handler
 * Purpose:  Update the color of the LEDs on execution of the timer 0 interrupt
 */
INTERRUPT(SIG_OUTPUT_COMPARE0A){
  uchar i;

  if(gc < TIMEOUT_WAIT){    /* Timeout Warning */
    for(i = 0; i < 3; i++){
      if(track[i].state){   /* State on */
	if(!track[i].count){
	  if((track[i].count = (0xf - track[i].color))){
	    PORTC |= track[i].ledbit;
	    track[i].state = STATE_OFF;
	  }
	}
      } else {
	if(!track[i].count){
	  if((track[i].count = track[i].color)){
	    PORTC &= ~track[i].ledbit;
	    track[i].state = STATE_ON;
	  }
	}
      }
      track[i].count--;
    }
  }

  /* If we don't get an update from the host, start blinking */
  if(initted){   /* Don't warn if we haven't been initialized */
    if(gc == TIMEOUT_WAIT) 
      PORTC |= (GREEN_BIT | RED_BIT | BLUE_BIT);  /* All off */
    else if(gc == WARN_ON)
      PORTC &= ~RED_BIT;                        /* Red on */
    else if(gc > WARN_OFF){
      gc = TIMEOUT_WAIT;
      return;
    }
    
    gc++;
  }
}

uint8_t eeprom_read(uint16_t addr){
  volatile uint8_t addrhi, addrlow;

  addrhi = ((addr >> 4) & 1);
  addrlow = addr & 0xff;

  /* Make sure we aren't writing */
  while(EECR & (1 << EEWE));

  EEARH = addrhi;
  EEARL = addrlow;

  EECR |= (1 << EERE);

  return EEDR;
}

void eeprom_write(uint16_t addr, uint8_t val){
  volatile uint8_t addrhi, addrlow;

  addrhi = ((addr >> 4) & 1);
  addrlow = addr & 0xff;

  /* Make sure we aren't writing */
  while(EECR & (1 << EEWE));

  EEARH = addrhi;
  EEARL = addrlow;
  EEDR = val;

  cli();   /* Make sure not to interrupt during write */

  EECR |= 1 << EEMWE;
  EECR |= 1 << EEWE;

  sei();
}

void timerInit(){
  TCCR0B = 5;		        /* set prescaler to 1/1024 */
  OCR0A = 8;                    /* Approximately 600/sec with above prescale */
  TCCR0A = 1 << WGM01;          /* ctc mode */
  TIMSK0 = 1 << OCIE0A;         /* interrupt on compare 0 A match */
}

void LEDInit(){
  uchar i;

  DDRC = 7;
  PORTC = 7;

  for(i = 0; i < 3; i++){
    track[i].color = 0;
    track[i].state = STATE_OFF;
    track[i].count = 0;
    track[i].ledbit = (1 << i);
  }
}

void eeprom_set_color(uchar *color){
  eeprom_write(0, color[0]);
  eeprom_write(1, color[1]);
}

void eeprom_get_color(){
  uchar color[2] = { 0, 0 };
  color[0] = eeprom_read(0);
  color[1] = eeprom_read(1);

  track[RED].color = (color[0] & 0xf);  
  track[GREEN].color = (color[1] >> 4); 
  track[BLUE].color = (color[1] & 0xf); 
}

static struct ledload_msg replymsg;

void send_color_default(){
  replymsg.command = REPLY_COLOR_DEFAULT;
  replymsg.color[0] = eeprom_read(0);
  replymsg.color[1] = eeprom_read(1);
}

uchar usb_request(uchar *data){
  uchar i;

  for(i = 0; i < sizeof(struct ledload_msg); i++)
    data[i] = ((uchar *)&replymsg)[i];
  
  return sizeof(struct ledload_msg);
}

uchar usbFunctionSetup(uchar stuff[8]){
    uchar command = stuff[1];
    uchar *data = &stuff[2];

    if(command == CHANGE_COLOR){
	/* Packet being sent need to be in the following order    */
	/* 0x0RGB on the host machine                             */
	/* The ordering changes because the bits on the MC aren't */
	/* the same                                               */
	track[RED].color = (data[0] & 0xf);         /* Green */
	track[GREEN].color = (data[1] >> 4);        /* Blue */
	track[BLUE].color = (data[1] & 0xf);        /* Red */
	
	gc = 0;
	initted = 1;
	
    } else if(command == CHANGE_COLOR_DEFAULT){
	eeprom_set_color(data);
	
    } else if(command == RESET_TO_DEFAULT){
	LEDInit();
	eeprom_get_color();
	
	initted = 0;
	gc = 0;
	
    } else if(command == REQ_COLOR_DEFAULT){
	send_color_default();    /* Prime the next IN request */

	usbMsgPtr = (uchar *)&replymsg;
    
	return sizeof(struct ledload_msg);
    }

    return 0;
}

uchar usbFunctionWrite(uchar *data, uchar len){
    return 0;
}

void usbReset(){
    LEDInit();
    eeprom_get_color();
    
    initted = 0;
    gc = 0;
}

/* Initialize and setup */
int main(void){
    LEDInit();
    
    usbInit();
    timerInit();
    
    sei();                        /* Turn on global interrupts */
    
    while(1){
	usbPoll();                /* Handle USB Messages */
    }
    
    return 0;
}

