/* 
   Copyright (c) Jeremy Garff 2005
   Please see included documentation for licensing information 
*/

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <unistd.h>
#include <termios.h>
#include <sys/io.h>
#include <errno.h>
#include <stdlib.h>

#include "../firmware/ledloadmsg.h"

int opendev(char *device){
    int fd;

    if((fd = open(device, O_RDWR)) < 0)
	return -1;

    return fd;
}

int reset_device(int fd){
  struct ledload_msg msg;

  msg.command = RESET_TO_DEFAULT;

  if(write(fd, &msg, sizeof(struct ledload_msg)) == -1)
    perror("write");

  return 0;
}

int change_default(int fd, unsigned char red, unsigned char green,
		   unsigned char blue){
  struct ledload_msg msg;

  if(red > 0xf)
    return -1;
  if(green > 0xf)
    return -1;
  if(blue > 0xf)
    return -1;

  msg.command = CHANGE_COLOR_DEFAULT;
  msg.color[0] = red;
  msg.color[1] = blue | (green << 4);

  if(write(fd, &msg, sizeof(struct ledload_msg)) == -1)
    perror("write");

  return 0;
}

int get_default(int fd, unsigned char *red, unsigned char *green,
		unsigned char *blue){
  struct ledload_msg msg;
  int r;

  msg.command = REQ_COLOR_DEFAULT;

  /* Send command */
  if(write(fd, &msg, sizeof(struct ledload_msg)) == -1)
    perror("write");

  /* Get response */
  r = read(fd, &msg, sizeof(struct ledload_msg));
  if(r != sizeof(struct ledload_msg))
    printf("read error; read %d bytes\n", r);

  if(msg.command != REPLY_COLOR_DEFAULT)
    printf("Incorrect reply command %d\n", msg.command);

  *red = msg.color[0] & 0xf;
  *green = msg.color[1] >> 4;
  *blue = msg.color[1] & 0xf;

  return 0;
}

int put_load(int fd, unsigned short byte){
  struct ledload_msg msg;
  unsigned short tmp;

  msg.command = CHANGE_COLOR;
  if(byte > 14){
      byte -= 15;
      tmp = byte;
      msg.color[1] = (0xf - tmp) << 4;  // Green
      msg.color[0] = tmp;               // Red
  } else {
      tmp = byte;
      msg.color[1] = (0xf - tmp);       // Blue
      msg.color[1] |= tmp << 4;         // Green
      msg.color[0] = 0;                 // Red
  }

  if(write(fd, &msg, sizeof(struct ledload_msg)) == -1){
    perror("write");
    return -1;
  }

  return 0;
}

/* Get the current statistics on the cpu */
int cpu(int pstat){
  unsigned int len, status;
  float diff;
  unsigned int total, idle, user, nice, system;
  unsigned int nidle, nuser, nnice, nsystem;
  static unsigned int last_idle, last_user, last_nice, last_system;
  char buf[5000], dummy[40];

  len = read(pstat, buf, sizeof(buf));
  status = sscanf(buf,"%s %d %d %d %d", dummy, &user, &nice, &system, &idle);

  nidle = idle - last_idle;
  nuser = user - last_user;
  nnice = nice - last_nice;
  nsystem = system - last_system;

  total = (nidle + nuser + nnice + nsystem);

  last_idle = idle;
  last_user = user;
  last_nice = nice;
  last_system = system;

  if(total != 0)
      diff = (total - nidle) / total;
  else 
      return 0;

  return(diff * 100);
}

#define RES 30

int loop(int fd){
  int pstat;
  int load, s = 0;
  int last = 0;

  if((pstat = open("/proc/stat", O_RDONLY)) == -1){
    printf("Unable to open /proc/stat");
    return -1;
  }
  
  while(1){
    lseek(pstat, 0, SEEK_SET);
    load = ((cpu(pstat) * RES) / 100);

    if(load == RES)
	s++;
    else if(!load)
	s--;
    else if(load > last)
	s++;
    else if(load < last)
	s--;

    last = load;

    if(s > RES)
	s = RES;
    if(s < 0)
	s = 0;

    if(put_load(fd, s) == -1)
      break;
    
    usleep(60000);
  }

  return(1);
}

void show_device_defaults(int fd){
  unsigned char red, green, blue;

  get_default(fd, &red, &green, &blue);

  printf("Red: %d Green: %d Blue: %d\n", red, green, blue);
}

void usage(char *name){
  printf("%s: <-g | -c | -r | -d [r] [g] [b]> <device>\n", name);
}

int main(int argv, char **argc){
  char dev[1024];
  pid_t id;
  int fd;

  switch(argv){
  case 3:
    sprintf(dev, "%s/load", argc[2]);
    if((fd = opendev(dev)) < 0){
      printf("Can't open device\n");
      return -1;
    }
    
    if(!strcmp(argc[1], "-c")){   /* Fork and do load monitor */
      if((id = fork()) == -1){
	printf("Fork error\n");
	return 0;
      }
      if(id==0){   
	setsid();
	loop(fd);
      } else 
	printf("CPU LED Load Monitor Started!\n"); 
      
    } else if(!strcmp(argc[1], "-r")){
      reset_device(fd);

    } else if(!strcmp(argc[1], "-g")){
      show_device_defaults(fd);

    } else 
      usage(argc[0]);

    break;

  case 6:
    sprintf(dev, "%s/load", argc[5]);
    if((fd = opendev(dev)) < 0){
      printf("Can't open device\n");
      return -1;
    }
    
    if(!strcmp(argc[1], "-d")){
      if(change_default(fd, atoi(argc[2]), atoi(argc[3]), atoi(argc[4])) == -1)
	printf("Color values need to be 0 - 15\n");
    } else
      usage(argc[0]);

    break;

  default:
    usage(argc[0]);

    break;
  } 

  return(0);
}



