/*
 * LEDLoad Device Driver
 *
 * Copyright (c) 2006 Jeremy Garff
 *
 * Author: Jeremy Garff
 *         http://www.jers.net/ledload
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/list.h>
#include <linux/spinlock.h>
#include <linux/leds.h>
#include <linux/module.h>
#include <linux/usb.h>
#include <asm/atomic.h>

#include "../firmware/ledloadmsg.h"

/* table of devices that work with this driver */
static struct usb_device_id id_table [] = {
    /* LEDLoad reserved USB device ID */
    { USB_DEVICE_VER(0x16c0, 0x460, 0x0, 0x22) },
    { },                             
};

struct usb_load {
    struct usb_device *udev;
    atomic_t lock;
    struct ledload_msg msg;
    unsigned char ledc0[2], ledc1[2];

    struct led_classdev redclass0;
    struct led_classdev greenclass0;
    struct led_classdev blueclass0;
};

MODULE_DEVICE_TABLE (usb, id_table);

static void submit_ledload_msg(struct usb_load *load){
    usb_control_msg(load->udev, usb_sndctrlpipe(load->udev, 0),
		    load->msg.command,
		    USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_DIR_OUT,
		    *(unsigned short *)&load->msg.color[0], 0, NULL, 0, 
		    USB_CTRL_SET_TIMEOUT);
}

static ssize_t send_ledload_msg(struct device *dev, struct device_attribute *attr,
                                const char *buf, size_t count){
  struct usb_interface *intf = to_usb_interface(dev);
  struct usb_load *load = usb_get_intfdata(intf);

  if(!load)
      return 0;

  if(atomic_read(&load->lock))
      return 0;

  if(count != sizeof(struct ledload_msg))
    return 0;

  atomic_set(&load->lock, 1);

  memcpy(&load->msg, buf, sizeof(struct ledload_msg));

  if(load->msg.command == CHANGE_COLOR){
      /* Save off the color sent to LED0 */
      load->ledc0[0] = load->msg.color[0];
      load->ledc0[1] = load->msg.color[1];
  }

#ifdef DEBUG
  printk("host -> device command %d color %04x\n", load->msg.command,
         *(unsigned short *)&load->msg.color);
#endif
  submit_ledload_msg(load);

  atomic_set(&load->lock, 0);

  return 0;
}

static int req_ledload_msg(struct usb_load *load){
    return usb_control_msg(load->udev, 
			   usb_rcvctrlpipe(load->udev, 0),
                           load->msg.command,
			   USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_DIR_OUT,
			   0, 0, &load->msg, sizeof(struct ledload_msg),
                           USB_CTRL_GET_TIMEOUT);
}

static ssize_t submit_ledload_req(struct device *dev, struct device_attribute *attr,
                                  char *buf){
    struct usb_interface *intf = to_usb_interface(dev);
    struct usb_load *load = usb_get_intfdata(intf);
    int ret;
    
    if(!load)
	return 0;
    
    if(atomic_read(&load->lock)){
	printk("ledload: device locked\n");
	return -1;
    }
    
    atomic_set(&load->lock, 1);
    
    ret = req_ledload_msg(load);
#ifdef DEBUG
    printk("device -> host ret %d command %d color %04x\n", ret,
           load->msg.command, *(unsigned short *)&load->msg.color);
#endif
    memcpy(buf, &load->msg, sizeof(struct ledload_msg));
    
    atomic_set(&load->lock, 0);
    
    return ret;
}

static DEVICE_ATTR(load, (S_IWUGO | S_IRUGO), submit_ledload_req, send_ledload_msg);

static void ledload_red_set(struct led_classdev *led_cdev, enum led_brightness value){
#ifdef NOTDEF
    struct usb_load *load = container_of(led_cdev, struct usb_load, redclass0);

    if(atomic_read(&load->lock))
	return;

    atomic_set(&load->lock, 1);

    load->msg.command = CHANGE_COLOR;
    load->msg.color[0] = load->ledc0[0];
    load->msg.color[1] = value >> 4;
    
    submit_ledload_msg(load);

    atomic_set(&load->lock, 0);
#endif
}

static void ledload_green_set(struct led_classdev *led_cdev, enum led_brightness value){
#ifdef NOTDEF
  struct usb_load *load = container_of(led_cdev, struct usb_load, greenclass0);

    if(atomic_read(&load->lock))
	return;

    atomic_set(&load->lock, 1);

    load->msg.command = CHANGE_COLOR;
    load->ledc0[0] &= 0xf0;
    load->msg.color[0] = load->ledc0[0] | (value >> 4);
    load->msg.color[1] = load->ledc0[1];
    
    submit_ledload_msg(load);

    atomic_set(&load->lock, 0);
#endif
}

static void ledload_blue_set(struct led_classdev *led_cdev, enum led_brightness value){
#ifdef NOTDEF
    struct usb_load *load = container_of(led_cdev, struct usb_load, blueclass0);

    if(atomic_read(&load->lock))
	return;

    atomic_set(&load->lock, 1);

    load->msg.command = CHANGE_COLOR;
    load->ledc0[0] &= 0x0f;
    load->msg.color[0] = load->ledc0[0] | ((value >> 4) << 4);
    load->msg.color[1] = load->ledc0[1];
    
    submit_ledload_msg(load);

    atomic_set(&load->lock, 0);
#endif
}

#ifdef CONFIG_PM
static int ledload_suspend(struct usb_interface *intf, pm_message_t state){
    struct usb_load *load = usb_get_intfdata(intf);

    led_classdev_suspend(&load->redclass0);    
    led_classdev_suspend(&load->blueclass0);
    led_classdev_suspend(&load->greenclass0);

    return 0;
}

static int ledload_resume(struct usb_interface *intf){
    struct usb_load *load = usb_get_intfdata(intf);

    led_classdev_resume(&load->redclass0);
    led_classdev_resume(&load->blueclass0);
    led_classdev_resume(&load->greenclass0);

    return 0;
}
#endif

static int ledload_probe(struct usb_interface *interface, 
			 const struct usb_device_id *id){
    struct ledload_msg msg;
    struct usb_device *udev;
    struct usb_load *dev;
    int ret = 0;

    udev = interface_to_usbdev(interface);

    if(!(dev = kmalloc(sizeof(struct usb_load), GFP_KERNEL)))
	return -ENOMEM;

    dev->udev = usb_get_dev(udev);

    atomic_set(&dev->lock, 0);
    memset(&dev->msg, 0, sizeof(struct ledload_msg));

    dev->redclass0.name = "ledload:red0";
    dev->redclass0.default_trigger = "none";
    dev->redclass0.brightness_set = ledload_red_set;

    dev->blueclass0.name = "ledload:blue0";
    dev->blueclass0.default_trigger = "none";
    dev->blueclass0.brightness_set = ledload_blue_set;

    dev->greenclass0.name = "ledload:green0";
    dev->greenclass0.default_trigger = "none";
    dev->greenclass0.brightness_set = ledload_green_set;

    usb_set_intfdata(interface, dev);

#ifdef NOTDEF
    /* Make the LEDs available to the LED driver */
    if((ret = led_classdev_register(&udev->dev, &dev->redclass0)) < 0){
	return ret;
    }
    
    if((ret = led_classdev_register(&udev->dev, &dev->greenclass0)) < 0){
	led_classdev_unregister(&dev->redclass0);
	return ret;
    }
    
    if((ret = led_classdev_register(&udev->dev, &dev->blueclass0)) < 0){
	led_classdev_unregister(&dev->redclass0);
	led_classdev_unregister(&dev->greenclass0);

	return ret;
    }
#endif

    msg.command = RESET_TO_DEFAULT;
    send_ledload_msg(&interface->dev, NULL, (char *)&msg, 
		     sizeof(struct ledload_msg));

    device_create_file(&interface->dev, &dev_attr_load);

    dev_info(&interface->dev, "LEDLoad Driver Started\n");

    return ret;
}

static void ledload_remove(struct usb_interface *interface){
    struct usb_load *dev;

    dev = usb_get_intfdata(interface);
    while(atomic_read(&dev->lock))
	msleep(10);

    device_remove_file(&interface->dev, &dev_attr_load);

    dev_info(&interface->dev, "LEDLoad Driver Stopped\n");

    usb_set_intfdata(interface, NULL);
    usb_put_dev(dev->udev);
    
//    led_classdev_unregister(&dev->redclass0);
//    led_classdev_unregister(&dev->greenclass0);
//    led_classdev_unregister(&dev->blueclass0);
    
    kfree(dev);
}

static struct usb_driver ledload_driver = {
	.probe		= ledload_probe,
	.disconnect     = ledload_remove,
#ifdef CONFIG_PM
	.suspend	= ledload_suspend,
	.resume		= ledload_resume,
#endif
	.name		= "ledload",
	.id_table       = id_table,
};

static int __init ledload_init(void){
    int ret = 0;

    if((ret = usb_register(&ledload_driver)))
	err("usb_register failed.  Error number %d\n", ret);

    return ret;
}

static void __exit ledload_exit(void){
    usb_deregister(&ledload_driver);
}

module_init(ledload_init);
module_exit(ledload_exit);

MODULE_AUTHOR("Jeremy Garff <jer @ jers.net>");
MODULE_DESCRIPTION("LEDLoad Driver");
MODULE_LICENSE("GPL");
